@@ -37,13 +37,14 @@ def client(env: str):
     for environment in environments:
         if environment["name"] == env:
             found = True
+            break
 
     if not found:
         raise NotFound(
             f"can't find environment {env}, create one by calling `cortex.cluster_client()`"
         )
 
-    return Client(env)
+    return Client(environment)
 
 
 def local_client(
@@ -85,38 +86,41 @@ def local_client(
 
 def cluster_client(
     name: str,
+    provider: str,
     operator_endpoint: str,
-    aws_access_key_id: str,
-    aws_secret_access_key: str,
+    aws_access_key_id: Optional[str] = None,
+    aws_secret_access_key: Optional[str] = None,
 ) -> Client:
     """
     Create a new environment to connect to an existing Cortex Cluster, and initialize a client to deploy and manage APIs on that cluster.
 
     Args:
         name: Name of the environment to create.
-        operator_endpoint: The endpoint for the operator of your Cortex Cluster. You can get this endpoint by running the CLI command `cortex cluster info`.
-        aws_access_key_id: AWS access key ID.
-        aws_secret_access_key: AWS secret access key.
+        provider: The provider of your Cortex cluster. Can be "aws" or "gcp".
+        operator_endpoint: The endpoint for the operator of your Cortex Cluster. You can get this endpoint by running the CLI command `cortex cluster info` for an AWS provider or `cortex cluster-gcp info` for a GCP provider.
+        aws_access_key_id: AWS access key ID. Required when `provider` is set to "aws".
+        aws_secret_access_key: AWS secret access key. Required when `provider` is set to "aws".
 
     Returns:
         Cortex client that can be used to deploy and manage APIs on a Cortex Cluster.
     """
-    run_cli(
-        [
-            "env",
-            "configure",
-            name,
-            "--provider",
-            "aws",
-            "--operator-endpoint",
-            operator_endpoint,
+    cli_args = [
+        "env",
+        "configure",
+        name,
+        "--provider",
+        provider,
+        "--operator-endpoint",
+        operator_endpoint,
+    ]
+    if provider == "aws":
+        cli_args += [
             "--aws-access-key-id",
             aws_access_key_id,
             "--aws-secret-access-key",
             aws_secret_access_key,
-        ],
-        hide_output=True,
-    )
+        ]
+    run_cli(cli_args, hide_output=True)
 
     return Client(name)
 
