@@ -22,47 +22,47 @@ import yaml
 import uuid
 import dill
 import inspect
+import shutil
 from pathlib import Path
 
 from typing import List, Dict, Optional, Tuple, Callable, Union
 from cortex.binary import run_cli, get_cli_path
 from cortex import util
 
+# Change if PYTHONVERSION changes
+EXPECTED_PYTHON_VERSION = "3.6.9"
+
 
 class Client:
-    def __init__(self, env: str):
+    def __init__(self, env: dict):
         """
         A client to deploy and manage APIs in the specified environment.
 
         Args:
-            env: Name of the environment to use.
+            env: Environment config
         """
         self.env = env
+        self.env_name = env["name"]
 
-    # CORTEX_VERSION_MINOR x5
-    def deploy(
+    # CORTEX_VERSION_MINOR
+    def create_api(
         self,
         api_spec: dict,
         predictor=None,
-        pip_dependencies=[],
-        conda_dependencies=[],
+        requirements=[],
+        conda_packages=[],
         project_dir: Optional[str] = None,
-        force: bool = False,
+        force: bool = True,
         wait: bool = False,
     ) -> list:
         """
         Deploy an API.
 
         Args:
-            api_spec: A dictionary defining a single Cortex API. Schema can be found here:
-                → Realtime API: https://docs.cortex.dev/v/0.23/deployments/realtime-api/api-configuration
-                → Batch API: https://docs.cortex.dev/v/0.23/deployments/batch-api/api-configuration
-                → Traffic Splitter: https://docs.cortex.dev/v/0.23/deployments/realtime-api/traffic-splitter
+            api_spec: A dictionary defining a single Cortex API. See https://docs.cortex.dev/v/0.24/ for schema.
             predictor: A Cortex Predictor class implementation. Not required when deploying a traffic splitter.
-                → Realtime API: https://docs.cortex.dev/v/0.23/deployments/realtime-api/predictors
-                → Batch API: https://docs.cortex.dev/v/0.23/deployments/batch-api/predictors
-            pip_dependencies: A list of PyPI dependencies that will be installed before the predictor class implementation is invoked.
-            conda_dependencies: A list of Conda dependencies that will be installed before the predictor class implementation is invoked.
+            requirements: A list of PyPI dependencies that will be installed before the predictor class implementation is invoked.
+            conda_packages: A list of Conda dependencies that will be installed before the predictor class implementation is invoked.
             project_dir: Path to a python project.
             force: Override any in-progress api updates.
             wait: Streams logs until the APIs are ready.
@@ -71,6 +71,11 @@ class Client:
             Deployment status, API specification, and endpoint for each API.
         """
 
+        if self.env["provider"] == "gcp" and wait:
+            raise ValueError(
+                "`wait` flag is not supported for clusters on GCP, please set the `wait` flag to false"
+            )
+
         if project_dir is not None and predictor is not None:
             raise ValueError(
                 "`predictor` and `project_dir` parameters cannot be specified at the same time, please choose one"
@@ -83,62 +88,67 @@ class Client:
                 yaml.dump([api_spec], f)  # write a list
                 return self._deploy(cortex_yaml_path, force, wait)
 
-        project_dir = Path.home() / ".cortex" / "deployments" / str(uuid.uuid4())
-        with util.open_tempdir(str(project_dir)):
-            cortex_yaml_path = os.path.join(project_dir, "cortex.yaml")
+        if api_spec.get("name") is None:
+            raise ValueError("`api_spec` must have the `name` key set")
 
-            if predictor is None:
-                # for deploying a traffic splitter
-                with open(cortex_yaml_path, "w") as f:
-                    yaml.dump([api_spec], f)  # write a list
-                    return self._deploy(cortex_yaml_path, force=force, wait=wait)
+        project_dir = Path.home() / ".cortex" / "deployments" / api_spec["name"]
 
-            # Change if PYTHONVERSION changes
-            expected_version = "3.6"
-            actual_version = f"{sys.version_info.major}.{sys.version_info.minor}"
-            if actual_version < expected_version:
-                raise Exception("cortex is only supported for python versions >= 3.6")  # unexpected
-            if actual_version > expected_version:
-                is_python_set = any(
-                    conda_dep.startswith("python=") or "::python=" in conda_dep
-                    for conda_dep in conda_dependencies
-                )
+        if project_dir.exists():
+            shutil.rmtree(str(project_dir))
 
-                if not is_python_set:
-                    conda_dependencies = [
-                        f"conda-forge::python={sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
-                    ] + conda_dependencies
+        project_dir.mkdir(parents=True)
 
-            if len(pip_dependencies) > 0:
-                with open(project_dir / "requirements.txt", "w") as requirements_file:
-                    requirements_file.write("\n".join(pip_dependencies))
-
-            if len(conda_dependencies) > 0:
-                with open(project_dir / "conda-packages.txt", "w") as conda_file:
-                    conda_file.write("\n".join(conda_dependencies))
-
-            if not inspect.isclass(predictor):
-                raise ValueError("predictor parameter must be a class definition")
-
-            with open(project_dir / "predictor.pickle", "wb") as pickle_file:
-                dill.dump(predictor, pickle_file)
-                if api_spec.get("predictor") is None:
-                    api_spec["predictor"] = {}
-
-                if predictor.__name__ == "PythonPredictor":
-                    predictor_type = "python"
-                if predictor.__name__ == "TensorFlowPredictor":
-                    predictor_type = "tensorflow"
-                if predictor.__name__ == "ONNXPredictor":
-                    predictor_type = "onnx"
-
-                api_spec["predictor"]["path"] = "predictor.pickle"
-                api_spec["predictor"]["type"] = predictor_type
+        cortex_yaml_path = os.path.join(project_dir, "cortex.yaml")
 
+        if predictor is None:
+            # for deploying a traffic splitter
             with open(cortex_yaml_path, "w") as f:
                 yaml.dump([api_spec], f)  # write a list
                 return self._deploy(cortex_yaml_path, force=force, wait=wait)
 
+        actual_version = (
+            f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
+        )
+
+        if actual_version != EXPECTED_PYTHON_VERSION:
+            is_python_set = any(
+                conda_dep.startswith("python=") or "::python=" in conda_dep
+                for conda_dep in conda_packages
+            )
+
+            if not is_python_set:
+                conda_packages = [f"python={actual_version}", "pip=19.*"] + conda_packages
+
+        if len(requirements) > 0:
+            with open(project_dir / "requirements.txt", "w") as requirements_file:
+                requirements_file.write("\n".join(requirements))
+
+        if len(conda_packages) > 0:
+            with open(project_dir / "conda-packages.txt", "w") as conda_file:
+                conda_file.write("\n".join(conda_packages))
+
+        if not inspect.isclass(predictor):
+            raise ValueError("predictor parameter must be a class definition")
+
+        with open(project_dir / "predictor.pickle", "wb") as pickle_file:
+            dill.dump(predictor, pickle_file)
+            if api_spec.get("predictor") is None:
+                api_spec["predictor"] = {}
+
+            if predictor.__name__ == "PythonPredictor":
+                predictor_type = "python"
+            if predictor.__name__ == "TensorFlowPredictor":
+                predictor_type = "tensorflow"
+            if predictor.__name__ == "ONNXPredictor":
+                predictor_type = "onnx"
+
+            api_spec["predictor"]["path"] = "predictor.pickle"
+            api_spec["predictor"]["type"] = predictor_type
+
+        with open(cortex_yaml_path, "w") as f:
+            yaml.dump([api_spec], f)  # write a list
+            return self._deploy(cortex_yaml_path, force=force, wait=wait)
+
     def _deploy(
         self,
         config_file: str,
@@ -161,9 +171,10 @@ class Client:
             "deploy",
             config_file,
             "--env",
-            self.env,
+            self.env_name,
             "-o",
             "mixed",
+            "-y",
         ]
 
         if force:
@@ -173,42 +184,48 @@ class Client:
 
         deploy_results = json.loads(output.strip())
 
+        deploy_result = deploy_results[0]
+
         if not wait:
-            return deploy_results
+            return deploy_result
+
+        # logging immediately will show previous versions of the replica terminating;
+        # wait a few seconds for the new replicas to start initializing
+        time.sleep(5)
 
         def stream_to_stdout(process):
             for c in iter(lambda: process.stdout.read(1), ""):
                 sys.stdout.write(c)
+                sys.stdout.flush()
 
-        for deploy_result in deploy_results:
-            api_name = deploy_result["api"]["spec"]["name"]
-            kind = deploy_result["api"]["spec"]["kind"]
-            if kind != "RealtimeAPI":
-                continue
+        api_name = deploy_result["api"]["spec"]["name"]
+        if deploy_result["api"]["spec"]["kind"] != "RealtimeAPI":
+            return deploy_result
 
-            env = os.environ.copy()
-            env["CORTEX_CLI_INVOKER"] = "python"
-            process = subprocess.Popen(
-                [get_cli_path(), "logs", "--env", self.env, api_name],
-                stderr=subprocess.STDOUT,
-                stdout=subprocess.PIPE,
-                encoding="utf8",
-                env=env,
-            )
+        env = os.environ.copy()
+        env["CORTEX_CLI_INVOKER"] = "python"
+        process = subprocess.Popen(
+            [get_cli_path(), "logs", "--env", self.env_name, api_name],
+            stderr=subprocess.STDOUT,
+            stdout=subprocess.PIPE,
+            encoding="utf8",
+            errors="replace",  # replace non-utf8 characters with `?` instead of failing
+            env=env,
+        )
 
-            streamer = threading.Thread(target=stream_to_stdout, args=[process])
-            streamer.start()
+        streamer = threading.Thread(target=stream_to_stdout, args=[process])
+        streamer.start()
 
-            while process.poll() is None:
-                api = self.get_api(api_name)
-                if api["status"]["status_code"] != "status_updating":
-                    if api["status"]["status_code"] == "status_live":
-                        time.sleep(2)
-                    process.terminate()
-                    break
-                time.sleep(2)
+        while process.poll() is None:
+            api = self.get_api(api_name)
+            if api["status"]["status_code"] != "status_updating":
+                time.sleep(5)  # accommodate latency in log streaming from the cluster
+                process.terminate()
+                break
+            time.sleep(5)
+        streamer.join(timeout=10)
 
-        return deploy_results
+        return api
 
     def get_api(self, api_name: str) -> dict:
         """
@@ -220,7 +237,7 @@ class Client:
         Returns:
             Information about the API, including the API specification, endpoint, status, and metrics (if applicable).
         """
-        output = run_cli(["get", api_name, "--env", self.env, "-o", "json"], hide_output=True)
+        output = run_cli(["get", api_name, "--env", self.env_name, "-o", "json"], hide_output=True)
 
         apis = json.loads(output.strip())
         return apis[0]
@@ -232,7 +249,7 @@ class Client:
         Returns:
             List of APIs, including information such as the API specification, endpoint, status, and metrics (if applicable).
         """
-        args = ["get", "-o", "json", "--env", self.env]
+        args = ["get", "-o", "json", "--env", self.env_name]
 
         output = run_cli(args, hide_output=True)
 
@@ -249,7 +266,7 @@ class Client:
         Returns:
             Information about the job, including the job status, worker status, and job progress.
         """
-        args = ["get", api_name, job_id, "--env", self.env, "-o", "json"]
+        args = ["get", api_name, job_id, "--env", self.env_name, "-o", "json"]
 
         output = run_cli(args, hide_output=True)
 
@@ -263,13 +280,35 @@ class Client:
             api_name: Name of the API to refresh.
             force: Override an already in-progress API update.
         """
-        args = ["refresh", api_name, "--env", self.env, "-o", "json"]
+        args = ["refresh", api_name, "--env", self.env_name, "-o", "json"]
 
         if force:
             args.append("--force")
 
         run_cli(args, hide_output=True)
 
+    def patch(self, api_spec: dict, force: bool = False) -> dict:
+        """
+        Update the api specification for an API that has already been deployed.
+
+        Args:
+            api_spec: The new api specification to apply
+            force: Override an already in-progress API update.
+        """
+
+        cortex_yaml_file = (
+            Path.home() / ".cortex" / "deployments" / f"cortex-{str(uuid.uuid4())}.yaml"
+        )
+        with util.open_temporarily(cortex_yaml_file, "w") as f:
+            yaml.dump([api_spec], f)
+            args = ["patch", cortex_yaml_file, "--env", self.env_name, "-o", "json"]
+
+            if force:
+                args.append("--force")
+
+            output = run_cli(args, hide_output=True)
+            return json.loads(output.strip())
+
     def delete_api(self, api_name: str, keep_cache: bool = False):
         """
         Delete an API.
@@ -282,7 +321,7 @@ class Client:
             "delete",
             api_name,
             "--env",
-            self.env,
+            self.env_name,
             "--force",
             "-o",
             "json",
@@ -306,7 +345,7 @@ class Client:
             api_name,
             job_id,
             "--env",
-            self.env,
+            self.env_name,
             "-o",
             "json",
         ]
@@ -323,7 +362,7 @@ class Client:
         Args:
             api_name: Name of the API.
         """
-        args = ["logs", api_name, "--env", self.env]
+        args = ["logs", api_name, "--env", self.env_name]
         run_cli(args)
 
     def stream_job_logs(
@@ -338,5 +377,5 @@ class Client:
             api_name: Name of the Batch API.
             job_id: Job ID.
         """
-        args = ["logs", api_name, job_id, "--env", self.env]
+        args = ["logs", api_name, job_id, "--env", self.env_name]
         run_cli(args)
